package me.parsell.glowstonewire.common;

import java.util.Map;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;

import org.jetbrains.annotations.Nullable;

import java.util.Iterator;

import net.minecraft.block.Block;
import net.minecraft.block.BlockState;
import net.minecraft.block.ShapeContext;
import net.minecraft.entity.EquipmentSlot;
import net.minecraft.entity.LivingEntity;
import net.minecraft.entity.player.PlayerEntity;
import net.minecraft.item.ItemPlacementContext;
import net.minecraft.item.ItemStack;
import net.minecraft.state.StateManager;
import net.minecraft.state.property.EnumProperty;
import net.minecraft.util.ActionResult;
import net.minecraft.util.BlockMirror;
import net.minecraft.util.BlockRotation;
import net.minecraft.util.Hand;
import net.minecraft.util.hit.BlockHitResult;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Direction;
import net.minecraft.util.shape.VoxelShape;
import net.minecraft.util.shape.VoxelShapes;
import net.minecraft.world.BlockView;
import net.minecraft.world.World;
import net.minecraft.world.WorldAccess;
import net.minecraft.world.WorldView;

public class GlowstoneWireBlock extends Block{
	// These are the states of the wires coming off our block
  public static final EnumProperty<GlowWireConnection> WIRE_CONNECTION_NORTH;
  public static final EnumProperty<GlowWireConnection> WIRE_CONNECTION_SOUTH;
  public static final EnumProperty<GlowWireConnection> WIRE_CONNECTION_EAST;
  public static final EnumProperty<GlowWireConnection> WIRE_CONNECTION_WEST;

	// Locally defined DIRECTION <-> WIRE_CONNECTION_DIRECTION
	public static final Map<Direction, EnumProperty<GlowWireConnection>> DIRECTION_TO_WIRE_CONNECTION;

	// Determines the hitbox for our block
	public static final VoxelShape DOT_VOXELSHAPE;
	private static final Map<Direction, VoxelShape> field_24414;
	private static final Map<Direction, VoxelShape> field_24415;


	public GlowstoneWireBlock(Settings settings) {
		super(settings);
		// Set these properties to be none as start, so we start with a dot
		this.setDefaultState((BlockState)((BlockState)((BlockState)((BlockState)((BlockState)((BlockState)this.stateManager.getDefaultState()).with(WIRE_CONNECTION_NORTH, GlowWireConnection.NONE)).with(WIRE_CONNECTION_EAST, GlowWireConnection.NONE)).with(WIRE_CONNECTION_SOUTH, GlowWireConnection.NONE)).with(WIRE_CONNECTION_WEST, GlowWireConnection.NONE)));
	}

	// Make sure out block has the properties for us to manage
	@Override
	protected void appendProperties(StateManager.Builder<Block, BlockState> builder) {
		builder.add(WIRE_CONNECTION_NORTH, WIRE_CONNECTION_EAST, WIRE_CONNECTION_SOUTH, WIRE_CONNECTION_WEST);
	}

	private static boolean hasConnection(BlockState state) {
		return ((GlowWireConnection)state.get(WIRE_CONNECTION_NORTH)).isConnected() || 
					((GlowWireConnection)state.get(WIRE_CONNECTION_SOUTH)).isConnected() ||
					((GlowWireConnection)state.get(WIRE_CONNECTION_EAST)).isConnected() ||
					((GlowWireConnection)state.get(WIRE_CONNECTION_WEST)).isConnected();
	}

	private static boolean hasUpConnection(BlockState state) {
		return ((GlowWireConnection)state.get(WIRE_CONNECTION_NORTH)).isConnectedUp() ||
				((GlowWireConnection)state.get(WIRE_CONNECTION_SOUTH)).isConnectedUp() ||
				((GlowWireConnection)state.get(WIRE_CONNECTION_EAST)).isConnectedUp() ||
				((GlowWireConnection)state.get(WIRE_CONNECTION_WEST)).isConnectedUp();
	}

	private BlockState determineState(WorldAccess world, BlockState state, BlockPos pos){
		Iterator<Direction> dirItr = Direction.Type.HORIZONTAL.iterator();
		while(dirItr.hasNext()){
			Direction dir = dirItr.next();
			state = state.with(DIRECTION_TO_WIRE_CONNECTION.get(dir), determineConnection(world, pos, dir));
		}
		return state;
	}

	private GlowWireConnection determineConnection(WorldAccess world, BlockPos pos, Direction direction){
		if((world.getBlockState(pos.offset(direction).up()).isOf(this) && world.isAir(pos.up()))) // If offset+up is me && if up is air
			if(!world.getBlockState(pos.offset(direction).up()).get(DIRECTION_TO_WIRE_CONNECTION.get(direction.getOpposite())).isBroken()) //if offset+up is not broken
				return GlowWireConnection.UP;
		if(world.getBlockState(pos.offset(direction).down()).isOf(this) && world.isAir(pos.offset(direction))) // if offset+down is me && if offset is air
			if(!world.getBlockState(pos.offset(direction).down()).get(DIRECTION_TO_WIRE_CONNECTION.get(direction.getOpposite())).isBroken()) //if offset+down is not broken
				return GlowWireConnection.SIDE;		
		if(world.getBlockState(pos.offset(direction)).isOf(this)) // if offset is me
			if(!world.getBlockState(pos.offset(direction)).get(DIRECTION_TO_WIRE_CONNECTION.get(direction.getOpposite())).isBroken()) // if offset is not broken
				return GlowWireConnection.SIDE;
		return GlowWireConnection.NONE;
	}

	// TODO: Fix this
	@Override
	public VoxelShape getOutlineShape(BlockState state, BlockView world, BlockPos pos, ShapeContext context) {
		VoxelShape voxelShape = DOT_VOXELSHAPE;
		Iterator var3 = Direction.Type.HORIZONTAL.iterator();
  
		while(var3.hasNext()) {
		   Direction direction = (Direction)var3.next();
		   GlowWireConnection wireConnection = state.get(DIRECTION_TO_WIRE_CONNECTION.get(direction));
		   if (wireConnection == GlowWireConnection.SIDE) {
			  voxelShape = VoxelShapes.union(voxelShape, (VoxelShape)field_24414.get(direction));
		   } else if (wireConnection == GlowWireConnection.UP) {
			  voxelShape = VoxelShapes.union(voxelShape, (VoxelShape)field_24415.get(direction));
		   }
		}
  
		return voxelShape;
	}

	public boolean canPlaceAt(BlockState state, WorldView world, BlockPos pos) {
		BlockPos blockPos = pos.down();
		BlockState blockState = world.getBlockState(blockPos);
		return this.canRunOnTop(world, blockPos, blockState);
	}

	private boolean canRunOnTop(BlockView world, BlockPos pos, BlockState floor) {
		return floor.isSideSolidFullSquare(world, pos, Direction.UP);
	}

	@Override
	public BlockState rotate(BlockState state, BlockRotation rotation) {
		switch(rotation) {
		case CLOCKWISE_180:
			 return (BlockState)((BlockState)((BlockState)((BlockState)state.with(WIRE_CONNECTION_NORTH, state.get(WIRE_CONNECTION_SOUTH))).with(WIRE_CONNECTION_EAST, state.get(WIRE_CONNECTION_WEST))).with(WIRE_CONNECTION_SOUTH, state.get(WIRE_CONNECTION_NORTH))).with(WIRE_CONNECTION_WEST, state.get(WIRE_CONNECTION_EAST));
		case COUNTERCLOCKWISE_90:
			 return (BlockState)((BlockState)((BlockState)((BlockState)state.with(WIRE_CONNECTION_NORTH, state.get(WIRE_CONNECTION_EAST))).with(WIRE_CONNECTION_EAST, state.get(WIRE_CONNECTION_SOUTH))).with(WIRE_CONNECTION_SOUTH, state.get(WIRE_CONNECTION_WEST))).with(WIRE_CONNECTION_WEST, state.get(WIRE_CONNECTION_NORTH));
		case CLOCKWISE_90:
			 return (BlockState)((BlockState)((BlockState)((BlockState)state.with(WIRE_CONNECTION_NORTH, state.get(WIRE_CONNECTION_WEST))).with(WIRE_CONNECTION_EAST, state.get(WIRE_CONNECTION_NORTH))).with(WIRE_CONNECTION_SOUTH, state.get(WIRE_CONNECTION_EAST))).with(WIRE_CONNECTION_WEST, state.get(WIRE_CONNECTION_SOUTH));
		default:
			 return state;
		}
	}

 	@Override
	 public BlockState mirror(BlockState state, BlockMirror mirror) {
		switch(mirror) {
		case LEFT_RIGHT:
			 return (BlockState)((BlockState)state.with(WIRE_CONNECTION_NORTH, state.get(WIRE_CONNECTION_SOUTH))).with(WIRE_CONNECTION_SOUTH, state.get(WIRE_CONNECTION_NORTH));
		case FRONT_BACK:
			 return (BlockState)((BlockState)state.with(WIRE_CONNECTION_EAST, state.get(WIRE_CONNECTION_WEST))).with(WIRE_CONNECTION_WEST, state.get(WIRE_CONNECTION_EAST));
		default:
			 return super.mirror(state, mirror);
		}
	}

	// Place block check disconnect (if block block connection)
 	@Override // For each link, set to break and update the blocks around me (above and below)
	public ActionResult onUse(BlockState state, World world, BlockPos pos, PlayerEntity player, Hand hand, BlockHitResult hit){
		if(player.abilities.allowModifyWorld)
			if(hasConnection(state) && !player.hasStackEquipped(EquipmentSlot.MAINHAND)) {
				Iterator<Direction> dirItr = Direction.Type.HORIZONTAL.iterator();
				while(dirItr.hasNext()){
					Direction dir = dirItr.next();
					if(state.get(DIRECTION_TO_WIRE_CONNECTION.get(dir)).isConnected())
						state = state.with(DIRECTION_TO_WIRE_CONNECTION.get(dir), GlowWireConnection.BROKE);
				}
				world.setBlockState(pos, state);
				// These call neighborUpdate not getStateForNeighborUpdate!!
				//world.updateNeighborsAlways(pos, this); // Update those around me
				world.updateNeighborsAlways(pos.down(), this); // Update those below me
				world.updateNeighborsAlways(pos.up(), this); // Update those above me
				return ActionResult.SUCCESS;
			}
		return ActionResult.PASS;
	}

	@Override // Update the link in the direction of the update, and if it's a block above me, break all links that are UP, down carry on as normal (i will break so dont do anything)
	public BlockState getStateForNeighborUpdate(BlockState state, Direction direction, BlockState newState, WorldAccess world, BlockPos pos, BlockPos posFrom) {
		System.out.println("State for Neighbor Update");
		if (direction == Direction.DOWN)
			return state; // Do nothing
		else if(direction == Direction.UP){
			if(!hasUpConnection(state))
				return state;
			Iterator<Direction> dirItr = Direction.Type.HORIZONTAL.iterator();
			while(dirItr.hasNext()){
				Direction dir = dirItr.next();
				if(!state.get(DIRECTION_TO_WIRE_CONNECTION.get(dir)).isBroken())
					state = state.with(DIRECTION_TO_WIRE_CONNECTION.get(dir), determineConnection(world, pos, dir));
			}
			return state;	
		} else if(state.get(DIRECTION_TO_WIRE_CONNECTION.get(direction)).isBroken())
			return state;
		return state.with(DIRECTION_TO_WIRE_CONNECTION.get(direction), determineConnection(world, pos, direction));
	}

	@Override // Get the state I should place in, and update those around me (above/below) as needed
	public BlockState getPlacementState(ItemPlacementContext ctx){
		System.out.println("Placement");
		return determineState(ctx.getWorld(), this.getDefaultState(), ctx.getBlockPos());
	}

	@Override
	public void onPlaced(World world, BlockPos pos, BlockState state, @Nullable LivingEntity placer, ItemStack itemStack) {
		Iterator<Direction> dirItr = Direction.Type.HORIZONTAL.iterator();
		while(dirItr.hasNext()){
			Direction dir = dirItr.next();
			if (state.get(DIRECTION_TO_WIRE_CONNECTION.get(dir)).isConnectedUp()) {
				BlockPos posOffset = pos.offset(dir).offset(Direction.UP);
				world.getBlockState(posOffset).neighborUpdate(world, posOffset, this, pos, false);
			}
		}
	}

	@Override // Called when I get told to update, so I should see what/why I need to update. I also use this to determine if I need to be drop or not.
	public void neighborUpdate(BlockState state, World world, BlockPos pos, Block block, BlockPos fromPos, boolean notify) {
		System.out.println(pos + " Neighbor Update from " + fromPos);
		if (!world.isClient) {
			if(!state.canPlaceAt(world, pos)) {
				dropStacks(state, world, pos);
				world.removeBlock(pos, false);
			} else {
				Iterator<Direction> dirItr = Direction.Type.HORIZONTAL.iterator();
				while(dirItr.hasNext()){
					Direction dir = dirItr.next();
					if(!state.get(DIRECTION_TO_WIRE_CONNECTION.get(dir)).isBroken())
						state = state.with(DIRECTION_TO_WIRE_CONNECTION.get(dir), determineConnection(world, pos, dir));
				}
				world.setBlockState(pos, state);
			}
		}
	}

	static {
		WIRE_CONNECTION_NORTH = EnumProperty.of("north", GlowWireConnection.class);
		WIRE_CONNECTION_EAST = EnumProperty.of("east", GlowWireConnection.class);
		WIRE_CONNECTION_SOUTH = EnumProperty.of("south", GlowWireConnection.class);
		WIRE_CONNECTION_WEST = EnumProperty.of("west", GlowWireConnection.class);
		
		DIRECTION_TO_WIRE_CONNECTION = Maps.newEnumMap((Map)ImmutableMap.of(Direction.NORTH, WIRE_CONNECTION_NORTH, Direction.EAST, WIRE_CONNECTION_EAST, Direction.SOUTH, WIRE_CONNECTION_SOUTH, Direction.WEST, WIRE_CONNECTION_WEST));
		
		DOT_VOXELSHAPE = Block.createCuboidShape(3.0D, 0.0D, 3.0D, 13.0D, 1.0D, 13.0D);
		field_24414 = Maps.newEnumMap((Map)ImmutableMap.of(Direction.NORTH, Block.createCuboidShape(3.0D, 0.0D, 0.0D, 13.0D, 1.0D, 13.0D), Direction.SOUTH, Block.createCuboidShape(3.0D, 0.0D, 3.0D, 13.0D, 1.0D, 16.0D), Direction.EAST, Block.createCuboidShape(3.0D, 0.0D, 3.0D, 16.0D, 1.0D, 13.0D), Direction.WEST, Block.createCuboidShape(0.0D, 0.0D, 3.0D, 13.0D, 1.0D, 13.0D)));
		field_24415 = Maps.newEnumMap((Map)ImmutableMap.of(Direction.NORTH, VoxelShapes.union((VoxelShape)field_24414.get(Direction.NORTH), Block.createCuboidShape(3.0D, 0.0D, 0.0D, 13.0D, 16.0D, 1.0D)), Direction.SOUTH, VoxelShapes.union((VoxelShape)field_24414.get(Direction.SOUTH), Block.createCuboidShape(3.0D, 0.0D, 15.0D, 13.0D, 16.0D, 16.0D)), Direction.EAST, VoxelShapes.union((VoxelShape)field_24414.get(Direction.EAST), Block.createCuboidShape(15.0D, 0.0D, 3.0D, 16.0D, 16.0D, 13.0D)), Direction.WEST, VoxelShapes.union((VoxelShape)field_24414.get(Direction.WEST), Block.createCuboidShape(0.0D, 0.0D, 3.0D, 1.0D, 16.0D, 13.0D))));
	}
}